import 'package:flutter/material.dart';
import 'package:flutter_requery/cache.dart';

typedef _QueryFuture<T> = Future<T> Function();

class QueryResponse<T> {
  T? data;
  bool loading;
  dynamic? error;

  QueryResponse({
    this.data,
    this.loading = false,
    this.error,
  });
  factory QueryResponse.initial() {
    return QueryResponse(loading: true);
  }
  factory QueryResponse.initialWithData(T? data) {
    return QueryResponse(data: data, loading: true);
  }
  factory QueryResponse.error(e) {
    return QueryResponse(error: e);
  }
  factory QueryResponse.data(T data) {
    return QueryResponse(data: data);
  }

  @override
  String toString() {
    return 'QueryResponse(data: ${data != null ? 'YES' : 'NO'}, loading: $loading, error: $error)';
  }
}

typedef _QueryBuilder = Widget Function(
  BuildContext context,
  dynamic data,
  bool loading,
  dynamic error,
);

class Query<T> extends StatefulWidget {
  final String cacheKey;
  final _QueryFuture<T> future;
  final _QueryBuilder builder;

  Query({
    required this.cacheKey,
    required this.builder,
    required this.future,
  });

  @override
  _QueryState<T> createState() => _QueryState<T>();
}

class _QueryState<T> extends State<Query<T>> {
  late Stream<QueryResponse<T>> stream;
  Stream<QueryResponse<T>> getActionStream<T>(
    Function cb,
  ) async* {
    try {
      T? data = cache.getData<T>(widget.cacheKey);
      yield QueryResponse.initialWithData(data);
      T response = await cb();
      cache.addData(widget.cacheKey, response);
      yield QueryResponse.data(response);
    } catch (e) {
      print(e);
      yield QueryResponse.error(e);
    }
  }

  Stream<QueryResponse<T>> getDataStream<T>(
    Function cb,
  ) async* {
    T data = cache.getData<T>(widget.cacheKey)!;
    yield QueryResponse.data(data);
  }

  @override
  void initState() {
    stream = cache.contains(widget.cacheKey)
        ? getDataStream(widget.future)
        : getActionStream(widget.future);
    cache.addSubscriber(
      widget.cacheKey,
      CacheItemSubscriber(
        id: widget.hashCode,
        notify: _rebuild,
      ),
    );
    super.initState();
  }

  @override
  void didUpdateWidget(covariant Query<T> oldWidget) {
    if (widget.cacheKey != oldWidget.cacheKey) {
      cache.removeSubscriber(oldWidget.cacheKey, oldWidget.hashCode);
      setState(() {
        stream = cache.contains(widget.cacheKey)
            ? getDataStream(widget.future)
            : getActionStream(widget.future);
        cache.addSubscriber(
          widget.cacheKey,
          CacheItemSubscriber(
            id: widget.hashCode,
            notify: _rebuild,
          ),
        );
      });
    }
    super.didUpdateWidget(oldWidget);
  }

  @override
  void dispose() {
    cache.removeSubscriber(widget.cacheKey, widget.hashCode);
    super.dispose();
  }

  void _rebuild() {
    if (mounted) {
      setState(() {
        stream = getActionStream(widget.future);
      });
    }
  }

  @override
  Widget build(BuildContext context) {
    return StreamBuilder(
      initialData: null,
      stream: stream,
      builder: (BuildContext context, AsyncSnapshot<QueryResponse?> snapshot) {
        if (snapshot.data == null) {
          return Container();
        }
        QueryResponse data = snapshot.data!;
        return this.widget.builder(
              context,
              data.data,
              data.loading,
              data.error,
            );
      },
    );
  }
}

void refetchQuery(String key) {
  cache.notifySubscribers(key);
}

void removeQuery(String key) {
  cache.removeSubscribers(key);
}

void resetStore() {
  cache.reset();
}
